/* wav2ambe.c */

// wav2ambe: convert wav-files to .ambe-files
// usage wav2ambe infile.wav outfile.ambe
// this program requires a DVdongle to do AMBE encoding, connected to
// a USB-port
// this program requires the sndfile-libraries to import wav-files
// see http://www.mega-nerd.com/libsndfile/

// The goal is to use this program to produce audio announcements on
// DSTAR repeaters by stitching pre-encoded AMBE files together

// copyright (C) 2010 Kristoff Bonne ON1ARF
/*
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

//  release info:
// 17 okt. 2010: version 0.0.1. Initial release
// 7 nov. 2010: version 0.1.0 Multithreaded release

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <stdint.h>

// sndfile: read wav-files
#include <sndfile.h>

// term io: serial port
#include <termios.h>

// posix threads + inter-process control
#include <pthread.h>
#include <signal.h>

// posix interupt timers
#include <time.h>



// change this to 1 for debugging
#define DEBUG 0



// local include-files
// wav2ambe.h: data structures + global data
// As wav2ambe is multithreaded, global data-structures are used
// by to threads to communicate with each other
#include "wav2ambe.h"

// serialsend.h: contains the function with that name
// Is started every 20 ms to send data to dongle
#include "serialsend.h"

// serialreceive.h: contains the function with that name
// Is started as a seperate thread by the main function
#include "serialreceive.h"


// defines for timed interrupts
#define CLOCKID CLOCK_REALTIME
#define SIG SIGRTMIN



//////////////////////////////////////////////////////////////////////////
// configuration part // // configuration part // // configuration part //
//////////////////////////////////////////////////////////////////////////
char * dongledevice="/dev/ttyUSB0";

int prepackets_total = 40;
int prepackets_saved = 10;

// end configuration part. Do not change below unless you
// know what you are doing


//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// Main program // Main program // Main program // Main program //////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////


int main(int argc,char** argv) {



// data structures:

// termios: structure to configure serial port
struct termios tio;

useconds_t waittime=5050; // 5050 µs = 1/4 of 20 ms sample-rate plus a little bit of extra waittime
useconds_t waittime_short=100; // 100 µs wait-time when loosing sync


// in-buffer for serial port: packets can be up to 8192 bytes
unsigned char serialinbuff[8192];

// SNDFILE: vars for library to read wave-files
SNDFILE * infile;
SF_INFO sfinfo;


// global data: globaldata-structure
globaldatastr globaldata;

// global data: structure for PCM-frame + related integers
int dataready;
// global data: other integers
int welive_serialreceive;


// posix threads
pthread_t thr_serialreceive;


// vars for timed interrupts
struct sigaction sa;
struct sigevent sev;
timer_t timerid;
struct itimerspec its;

// some miscellaneous vars:
int status;
int stop;
int ret;
int frametype;
int dummysend;
int framecount;

char * formatin;

int prepackets_first;

prepackets_first=prepackets_total - prepackets_saved;


// ////// data definition done ///


// main program starts here ///
// part 1: initialise vars and check cli-arguments


// start of program //

if (argc < 4) {
	fprintf(stderr,"Error: at least 3 arguments needed.\n");
	fprintf(stderr,"usage: %s {a,d} infile.wav outfile.ambe\n",argv[0]);
	exit(-1);
}; // end if


// verify format: 1st argument
// can be "a" for .ambe files, or "d" for .dvtool files

outputfileformat=0;
formatin=argv[1];

// check first character, if 'a' or 'A' -> format is 1
if ((formatin[0] == 'a') || (formatin[0] == 'A')) {
	outputfileformat=1;
}; // end if

// check first character, if 'd' or 'D' -> format is 2
if ((formatin[0] == 'd') || (formatin[0] == 'D')) {
	outputfileformat=2;
}; // end if

if (outputfileformat == 0) {
	fprintf(stderr,"Error: 1st argument must be 'a' for .ambe file-format or 'd' for .dvtool file-format\n");
	fprintf(stderr,"usage: %s {a,d} infile.wav outfile.ambe\n",argv[0]);
	exit(-1);
}; // end if

// copy name outfile to global var "fname_out"
fname_out=argv[3];

// initialise pcmbuffer: always 320 bytes (20 ms, 8000 sps, 16 bit/sample)
// buffsize (0x8142) actualy contains 2 data-items:
// - first 5 bits: "ptype" (here: 10000)
// - remaining 13 bites: length: 0x142 = 322 (320 databytes + 2 header bytes)
pcmbuffer.buffsize1=0x42;
pcmbuffer.buffsize2=0x81;
memset(&pcmbuffer.pcm,0,320);

// init global data
globaldata.p_welive_serialreceive = &welive_serialreceive;
globaldata.p_allpacketsend=&allpacketsend;


// open serial port to dongle
donglefd=open(dongledevice, O_RDWR | O_NONBLOCK);

if (donglefd == 0) {
	fprintf(stderr,"open failed! \n");
	exit(-1);
}; // end if

// store donglefd in globaldata, to be send to serialreceive tread lateron
globaldata.donglefd=donglefd;


// now configure serial port
// Serial port programming for Linux, see termios.h for more info
memset(&tio,0,sizeof(tio));
tio.c_iflag=0; tio.c_oflag=0; tio.c_cflag=CS8|CREAD|CLOCAL; // 8n1
tio.c_lflag=0; tio.c_cc[VMIN]=1; tio.c_cc[VTIME]=0;

cfsetospeed(&tio,B230400); // 230 Kbps
cfsetispeed(&tio,B230400); // 230 Kbps
tcsetattr(donglefd,TCSANOW,&tio);



// open wave file
if (! (infile = sf_open (argv[2],SFM_READ,&sfinfo))) {
	fprintf(stderr,"Unable to open input file! %s\n",argv[2]);
	exit(-1);
}; // end if


// file must be wave-file, 8000 Khz sampling, mono, 16 bit little-endian

if (sfinfo.samplerate != 8000) {
	fprintf(stderr,"Error: wav-file must be 8 Khz sampling.\n");
	exit(-1);
}; // end if

if (sfinfo.channels != 1) {
	fprintf(stderr,"Error: wav-file must mono.\n");
	exit(-1);
}; // end if

if (sfinfo.format != 0x10002) {
	fprintf(stderr,"Error: wav-file must MS-WAV, 16 bit Little-endian.\n");
	exit(-1);
}; // end if




// init some vars
status=0;


// First part of the program: initialise the dongle
// This runs pure sequencial

// main loop:
// senario:
// status 0: send "give-name" command to dongle
// status 1: wait for name -> send "start" command
// status 2: wait for confirmation of "start" -> send ambe-packet (configures encoder) + dummy pcm-frame


while (status < 3) {
	// some local vars
	int totalreceived;

	uint16_t plen_i; 
	uint16_t ptype;

if (DEBUG == 1) {
	fprintf(stderr,"status = %d\n",status);
}
	// read next frame, no error-correction, 2 ms of delay if no data
	// timeout is 4 seconds (2 ms * 2000) in status 3 (after sending
	// first command to dongle) or 1 (2 ms * 500) seconds otherwize
	if (status == 1) {
		ret=serialframereceive(donglefd,serialinbuff,&ptype,&plen_i,0,2000,2000);
	} else {
		ret=serialframereceive(donglefd,serialinbuff,&ptype,&plen_i,0,2000,500);
	}; // end else - if


// part 1: what if no data received ???	
	if (ret < 0) {
	// nothing received.

		// check for status
		if (status == 1) {
			// 4 second timeout after first command to dongle.
			// are you sure there is a dongle present?
			fprintf(stderr,"Error: timeout receiving data from dongle.! Exiting\n");
			exit(-1);
		}; // end if


		if (status == 0) {
		// status is 2 (after 2 seconds of waiting, send "request name")
if (DEBUG == 1) {
			printf("sending request name\n");
}; // end DEBUG
			ret=write (donglefd,request_name,4);
			// move to status 1
			status=1;
		}; // end if

	} else {
	// part2: data is received

if (DEBUG == 1) {
		printf("plen_i = %d, ptype = %d \n",plen_i, ptype);
}; // end DEBUG

if (DEBUG == 1) {
	int loop;
		// dump packet
		for (loop=0;loop<totalreceived;loop++) {
			printf("%x ",serialinbuff[loop]);
		}; // end for
		printf("\n");
}; // end DEBUG



		// parse received packet:
		// case: status 2 in senario
		// reply to start received -> send initial ambe-packet
		if ((status == 2) && (plen_i == 3) && (ptype == 0)) {
if (DEBUG == 1) {
		printf("sending initial ambe-frame\n");
}; // end DEBUG
			// also send a dummy pcm-frame
			ret=write (donglefd,(void *) &pcmbuffer,322);

			ret=write (donglefd,ambe,50);

			// move to status 3
			status=3;
		}; // end if

		// case: status 1 in senario:
		// dongle name received -> send "start" command
		if ((status == 1) && (plen_i == 12) && (ptype == 0) && (serialinbuff[0] == 1) && (serialinbuff[1] == 0)) {
if (DEBUG == 1) {
			printf("sending start\n");
}; // end DEBUG
			ret=write (donglefd,command_start,5);
			// move to status 2
			status=2;
		}; // end if


	}; // end else - if (data received or not?)

}; // end while (status < 3)


// Part 2 of main program

// now we are at status level 3. We have just received the acknowledge
// to the "start" and have send the 1st ambe-packet to configure
// the encoder

// now we do three things:
// - the main process reads pcm-frames of the .wav-file, and
// makes them available for the "send data" thread
// - the "send data" function is started via a interrupt-time every
// 20 ms, reads the data prepared by the main thread and sends it
// out to the dvdongle
// - we start the "serialreceive" thread that receives the frames from
// the dvdongle and saves the ambe-frames it to a file
// in addition to this, it also send a dummy ambe-frame when receiving
// a ambe-frame from the dongle

// init and start thread serialreceive
// copy fileformat parameter from global configuration
globaldata.fileformat=outputfileformat;

pthread_create(&thr_serialreceive,NULL,funct_serialreceive, (void *) &globaldata);

// init vars for serialsend function
datatosend=0;
datasendready=1;

// establing handler for signal 
sa.sa_flags = 0;
sa.sa_handler = funct_serialsend;
sigemptyset(&sa.sa_mask);

ret=sigaction(SIG, &sa, NULL);
if (ret <0) {
	fprintf(stderr,"error in sigaction!\n");
	exit;
}; // end if


/* Create the timer */
sev.sigev_notify = SIGEV_SIGNAL;
sev.sigev_signo = SIG;
sev.sigev_value.sival_ptr = &timerid;

ret=timer_create(CLOCKID, &sev, &timerid);
if (ret < 0) {
	fprintf(stderr,"error in timer_create!\n");
	exit;
}; // end if


// start timed function, every 20 ms
its.it_value.tv_sec = 0;
its.it_value.tv_nsec = 5000000; // 5 ms: start after 1/4 of first frame
its.it_interval.tv_sec = its.it_value.tv_sec;
its.it_interval.tv_nsec = 20000000; // 20 ms = 20 million nanoseconds

ret=timer_settime(timerid, 0, &its, NULL);
if (ret < 0) {
	fprintf(stderr,"error in timer_settime!\n");
	exit;
}; // end if


framecount=0;

// read data from wav-file
stop=0;

while (stop == 0) {

	// can we send the next frame?
	if (datasendready == 1) {
	// yes, data is send by serialsend, so let's prepare the next packet
		sf_count_t nsample;

		nsample=sf_read_short(infile,(short *) &pcmbuffer.pcm,160);

		// is there still data in the wav-file
		if (nsample >0) {
			// yes

			// if less then 160 samples received, all rest as all-0 and
			// set the "lastframe" flag
			if (nsample < 160) {
				// fill rest with all 0
				memset(&pcmbuffer.pcm[nsample*2],0,(160-nsample)*2);

				// set data-to-send marker to trigger serialsend function
				datatosend=1;
				datasendready=0;
				framecount++;

			} else {
				// this is strictly not necessairy, just to be sure
				allpacketsend=0;

				// set data-to-send marker to trigger serialsend function
				datatosend=1;
				datasendready=0;
				framecount++;
			}; // end else - if (complete frame?)
		} else {
			// no data to send anymore. End-of-file reached of wav-file
			allpacketsend=1;

			// set "stop" as we have finished reading the wav-file
			stop=1;
		}; // end else - if (still data in wav-file?)

	} else {
	// no data to send -> sleep for 2 ms
		usleep(2000);
	}; // end else - if (data send ready?)


}; // end while (stop)


// OK, we have read all of the file, now wait for the "serialsend" function to terminate
// this is done by monitoring the "allpacketsend", which is set to 3 when serialsend sends last frame

while (allpacketsend <2) {
	// not yet done, wait 5 ms
	usleep(5000);
}; // end while

// no stop interupt timer
its.it_value.tv_sec = 0;
its.it_value.tv_nsec = 0; // timer = 0 -> stop timer
its.it_interval.tv_sec = its.it_value.tv_sec;
its.it_interval.tv_nsec = its.it_value.tv_nsec;

ret=timer_settime(timerid, 0, &its, NULL);
if (ret < 0) {
	fprintf(stderr,"error in timer_settime!\n");
	exit;
}; // end if


// now wait for the serialreceive program to terminate
while (welive_serialreceive == 1) {
	// not yet done, wait 5 ms
	usleep(5000);
}; // end while


// OK, we are finished
printf("done: %d frames encoded\n", framecount);
close(donglefd);

}; // end main

